# -*- coding: utf-8 -*-

from django.db import models
from django.utils.translation import gettext_lazy as _


class Watermark(models.Model):

    name = models.CharField(max_length=50, verbose_name=_("name"))
    image = models.ImageField(upload_to="watermarks", verbose_name=_("image"))
    is_active = models.BooleanField(default=True, blank=True, verbose_name=_("is active"))

    # for internal use...

    date_created = models.DateTimeField(auto_now_add=True)
    date_updated = models.DateTimeField(auto_now=True)

    class Meta:
        ordering = ["name"]
        verbose_name = _("watermark")
        verbose_name_plural = _("watermarks")

    def __str__(self):
        return self.name

    def save(self, *args, **kwargs):
        if self.is_active:
            # select all other active items
            qs = self.__class__.objects.filter(name__exact=self.name, is_active=True)
            # except self (if self already exists)
            if self.pk:
                qs = qs.exclude(pk=self.pk)
            # and deactive them
            qs.update(is_active=False)

        super(Watermark, self).save(*args, **kwargs)



# Generated by Django 3.0.13 on 2021-03-20 18:45

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("watermarker", "0001_initial"),
    ]

    operations = [
        migrations.AlterField(
            model_name="watermark",
            name="image",
            field=models.ImageField(upload_to="watermarks", verbose_name="image"),
        ),
        migrations.AlterField(
            model_name="watermark",
            name="is_active",
            field=models.BooleanField(blank=True, default=True, verbose_name="is active"),
        ),
    ]



# -*- coding: utf-8 -*-


from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = []

    operations = [
        migrations.CreateModel(
            name="Watermark",
            fields=[
                ("id", models.AutoField(verbose_name="ID", serialize=False, auto_created=True, primary_key=True)),
                ("name", models.CharField(max_length=50, verbose_name="name")),
                ("image", models.ImageField(upload_to=b"watermarks", verbose_name="image")),
                ("is_active", models.BooleanField(default=True, verbose_name="is active")),
                ("date_created", models.DateTimeField(auto_now_add=True)),
                ("date_updated", models.DateTimeField(auto_now=True)),
            ],
            options={
                "ordering": ["name"],
                "verbose_name": "watermark",
                "verbose_name_plural": "watermarks",
            },
        ),
    ]



# -*- coding: utf-8 -*-

from appconf import AppConf
from django.conf import settings  # pylint: disable=W0611


class WatermarkSettings(AppConf):
    QUALITY = 85
    OBSCURE_ORIGINAL = True
    RANDOM_POSITION_ONCE = True

    class Meta:
        prefix = "watermark"
        holder = "watermarker.conf.settings"



# -*- coding: utf-8 -*-

import os

from django.test import TestCase
from PIL import Image

from ..utils import watermark


class UtilsTestCase(TestCase):
    def setUp(self):
        self.im = Image.open(os.path.join(os.path.dirname(__file__), "test.png"))
        self.mark = Image.open(os.path.join(os.path.dirname(__file__), "overlay.png"))

    def test_tile(self):
        watermark(self.im, self.mark, tile=True, opacity=0.5, rotation=30).save(
            os.path.join(os.path.dirname(__file__), "test1.png")
        )

    def test_scale(self):
        watermark(self.im, self.mark, scale="F").save(os.path.join(os.path.dirname(__file__), "test2.png"))

    def test_grayscale(self):
        watermark(self.im, self.mark, position=(100, 100), opacity=0.5, greyscale=True, rotation=-45).save(
            os.path.join(os.path.dirname(__file__), "test3.png")
        )

    def test_position(self):
        watermark(self.im, self.mark, position="C", tile=False, opacity=0.2, scale=2, rotation=30).save(
            os.path.join(os.path.dirname(__file__), "test4.png")
        )



# -*- coding: utf-8 -*-

from django.contrib import admin

from .models import Watermark


class WatermarkAdmin(admin.ModelAdmin):
    list_display = ["name", "is_active"]
    list_filter = ["is_active"]
    search_fields = ["name"]


admin.site.register(Watermark, WatermarkAdmin)



# -*- coding: utf-8 -*-
"""
Utilities for applying a watermark to an image using PIL.

Stolen from http://code.activestate.com/recipes/362879/

"""
import re
import random

from PIL import Image, ImageEnhance

from .conf import settings


def _percent(var):
    """
    Just a simple interface to the _val function with a more meaningful name.
    """
    return _val(var, True)


def _int(var):
    """
    Just a simple interface to the _val function with a more meaningful name.
    """
    return _val(var)


def _val(var, is_percent=False):
    """
    Tries to determine the appropriate value of a particular variable that is
    passed in.  If the value is supposed to be a percentage, a whole integer
    will be sought after and then turned into a floating point number between
    0 and 1.  If the value is supposed to be an integer, the variable is cast
    into an integer.

    """
    try:
        if is_percent:
            var = float(int(var.strip("%")) / 100.0)
        else:
            var = int(var)
    except ValueError:
        raise ValueError("invalid watermark parameter: " + var)
    return var


def reduce_opacity(img, opacity):
    """
    Returns an image with reduced opacity.
    """
    assert opacity >= 0 and opacity <= 1

    if img.mode != "RGBA":
        img = img.convert("RGBA")
    else:
        img = img.copy()

    alpha = img.split()[3]
    alpha = ImageEnhance.Brightness(alpha).enhance(opacity)
    img.putalpha(alpha)

    return img


def determine_scale(scale, img, mark):
    """
    Scales an image using a specified ratio, 'F' or 'R%%'. If `scale` is
    'F', the image is scaled to be as big as possible to fit in `img`
    without falling off the edges.  If `scale` is 'R%%', the watermark
    resizes to a percentage of minimum size of source image.  Returns
    the scaled `mark`.

    """
    if scale:
        try:
            scale = float(scale)
        except (ValueError, TypeError):
            pass

        if isinstance(scale, str) and scale.upper() == "F":
            # scale watermark to full, but preserve the aspect ratio
            scale = min(float(img.size[0]) / mark.size[0], float(img.size[1]) / mark.size[1])
        elif isinstance(scale, str) and re.match(r"R\d{1,3}\%", scale.upper()):
            # scale watermark to % of source image and preserve the aspect ratio
            percentage = float(re.match(r"R(\d{1,3})\%", scale.upper()).group(1))
            scale = (
                min(float(img.size[0]) / mark.size[0], float(img.size[1]) / mark.size[1])
                / 100 * percentage
            )
        elif not isinstance(scale, (float, int)):
            raise ValueError(
                'Invalid scale value "%s"! Valid values are "F" '
                'for ratio-preserving scaling, "R%%" for percantage aspect '
                "ratio of source image and floating-point numbers and "
                "integers greater than 0." % scale
            )

        # determine the new width and height
        w = int(mark.size[0] * float(scale))
        h = int(mark.size[1] * float(scale))

        # apply the new width and height, and return the new `mark`
        return (w, h)
    else:
        return mark.size


def determine_rotation(rotation, mark):
    """
    Determines the number of degrees to rotate the watermark image.
    """
    if isinstance(rotation, str) and rotation.lower() == "r":
        rotation = random.randint(0, 359)
    else:
        rotation = _int(rotation)

    return rotation


def determine_position(position, img, mark):
    """
    Options:
        TL: top-left
        TR: top-right
        BR: bottom-right
        BL: bottom-left
        C: centered
        R: random
        X%xY%: relative positioning on both the X and Y axes
        X%xY: relative positioning on the X axis and absolute positioning on the
              Y axis
        XxY%: absolute positioning on the X axis and relative positioning on the
              Y axis
        XxY: absolute positioning on both the X and Y axes

    """
    left = top = 0

    max_left = max(img.size[0] - mark.size[0], 0)
    max_top = max(img.size[1] - mark.size[1], 0)

    if not position:
        position = "r"

    if isinstance(position, tuple):
        left, top = position
    elif isinstance(position, str):
        position = position.lower()

        # corner positioning
        if position in ["tl", "tr", "br", "bl"]:
            if "t" in position:
                top = 0
            elif "b" in position:
                top = max_top
            if "l" in position:
                left = 0
            elif "r" in position:
                left = max_left

        # center positioning
        elif position == "c":
            left = int(max_left / 2)
            top = int(max_top / 2)

        # random positioning
        elif position == "r":
            left = random.randint(0, max_left)
            top = random.randint(0, max_top)

        # relative or absolute positioning
        elif "x" in position:
            left, top = position.split("x")

            if "%" in left:
                left = max_left * _percent(left)
            else:
                left = _int(left)

            if "%" in top:
                top = max_top * _percent(top)
            else:
                top = _int(top)

    return int(left), int(top)


def watermark(
    img,
    mark,
    position=(0, 0),
    opacity=1,
    scale=1.0,
    tile=False,
    greyscale=False,
    rotation=0,
    return_name=False,
    **kwargs
):
    """Adds a watermark to an image"""

    if opacity < 1:
        mark = reduce_opacity(mark, opacity)

    if not isinstance(scale, tuple):
        scale = determine_scale(scale, img, mark)

    if scale[0] != mark.size[0] and scale[1] != mark.size[1]:
        mark = mark.resize(scale, resample=Image.ANTIALIAS)

    if greyscale and mark.mode != "LA":
        mark = mark.convert("LA")

    rotation = determine_rotation(rotation, mark)
    if rotation != 0:
        # give some leeway for rotation overlapping
        new_w = int(mark.size[0] * 1.5)
        new_h = int(mark.size[1] * 1.5)

        new_mark = Image.new("RGBA", (new_w, new_h), (0, 0, 0, 0))

        # center the watermark in the newly resized image
        new_l = int((new_w - mark.size[0]) / 2)
        new_t = int((new_h - mark.size[1]) / 2)
        new_mark.paste(mark, (new_l, new_t))

        mark = new_mark.rotate(rotation)

    position = determine_position(position, img, mark)

    if img.mode != "RGBA":
        img = img.convert("RGBA")

    # make sure we have a tuple for a position now
    assert isinstance(position, tuple), 'Invalid position "%s"!' % position

    # create a transparent layer the size of the image and draw the
    # watermark in that layer.
    layer = Image.new("RGBA", img.size, (0, 0, 0, 0))
    if tile:
        first_y = int(position[1] % mark.size[1] - mark.size[1])
        first_x = int(position[0] % mark.size[0] - mark.size[0])

        for y in range(first_y, img.size[1], mark.size[1]):
            for x in range(first_x, img.size[0], mark.size[0]):
                layer.paste(mark, (x, y))
    else:
        layer.paste(mark, position)

    # composite the watermark with the layer
    return Image.composite(layer, img, layer)



# -*- coding: utf-8 -*-

import errno
import hashlib
import logging
import os
import traceback
from datetime import datetime

from PIL import Image

from urllib.parse import unquote
from urllib.request import url2pathname

from django import template
from django.utils.encoding import smart_str
from django.utils.timezone import get_default_timezone, is_aware, make_aware

from watermarker import utils
from watermarker.conf import settings
from watermarker.models import Watermark

QUALITY = settings.WATERMARK_QUALITY
OBSCURE_ORIGINAL = settings.WATERMARK_OBSCURE_ORIGINAL
RANDOM_POSITION_ONCE = settings.WATERMARK_RANDOM_POSITION_ONCE

register = template.Library()

logger = logging.getLogger("watermarker")


class Watermarker(object):
    def __call__(
        self,
        url,
        name,
        position=None,
        opacity=0.5,
        tile=False,
        scale=1.0,
        greyscale=False,
        rotation=0,
        noalpha=True,
        quality=QUALITY,
        obscure=OBSCURE_ORIGINAL,
        random_position_once=RANDOM_POSITION_ONCE,
    ):
        """
        Creates a watermarked copy of an image.
        """
        # look for the specified watermark by name.  If it's not there, go no
        # further
        try:
            watermark = Watermark.objects.get(name__exact=name, is_active=True)
        except Watermark.DoesNotExist:
            logger.error('Watermark "%s" does not exist... Bailing out.' % name)
            return url

        # make sure URL is a string
        url = smart_str(url)

        basedir = "%s/watermarked/" % os.path.dirname(url)
        original_basename, ext = os.path.splitext(os.path.basename(url))

        # open the target image file along with the watermark image
        target = Image.open(self._get_filesystem_path(url))
        mark = Image.open(watermark.image.path)

        # determine the actual value that the parameters provided will render
        random_position = bool(position is None or str(position).lower() == "r")
        scale = utils.determine_scale(scale, target, mark)
        mark = mark.resize(scale, resample=Image.ANTIALIAS)
        rotation = utils.determine_rotation(rotation, mark)
        pos = utils.determine_position(position, target, mark)

        # see if we need to create only one randomly positioned watermarked
        # image
        if not random_position or (not random_position_once and random_position):
            logger.debug("Generating random position for watermark each time")
            position = pos
        else:
            logger.debug("Random positioning watermark once")

        params = {
            "position": position,
            "opacity": opacity,
            "scale": scale,
            "tile": tile,
            "greyscale": greyscale,
            "rotation": rotation,
            "original_basename": original_basename,
            "ext": ext,
            "noalpha": noalpha,
            "quality": quality,
            "watermark": watermark.id,
            "left": pos[0],
            "top": pos[1],
            "fstat": os.stat(self._get_filesystem_path(url)),
        }
        logger.debug("Params: %s" % params)

        fname = self.generate_filename(mark, **params)
        url_path = self.get_url_path(basedir, original_basename, ext, fname, obscure)
        fpath = self._get_filesystem_path(url_path)

        logger.debug(
            "Watermark name: %s; URL: %s; Path: %s"
            % (
                fname,
                url_path,
                fpath,
            )
        )

        # see if the image already exists on the filesystem. If it does, use it.
        if os.access(fpath, os.R_OK):
            # see if the ``Watermark`` object was modified since the
            # file was created
            modified = make_aware(datetime.fromtimestamp(os.path.getmtime(fpath)), get_default_timezone())
            date_updated = watermark.date_updated
            if not is_aware(date_updated):
                date_updated = make_aware(date_updated, get_default_timezone())
            # only return the old file if things appear to be the same
            if modified >= date_updated:
                logger.info("Watermark exists and has not changed. Bailing out.")
                return url_path

        # make sure the position is in our params for the watermark
        params["position"] = pos

        self.create_watermark(target, mark, fpath, **params)

        # send back the URL to the new, watermarked image
        return url_path

    def _get_filesystem_path(self, url_path, basedir=settings.MEDIA_ROOT):
        """Makes a filesystem path from the specified URL path"""

        if url_path.startswith(settings.MEDIA_URL):
            url_path = url_path[len(settings.MEDIA_URL):]  # strip media root url

        return os.path.normpath(os.path.join(basedir, url2pathname(url_path)))

    def generate_filename(self, mark, **kwargs):
        """Comes up with a good filename for the watermarked image"""

        kwargs = kwargs.copy()

        kwargs["opacity"] = int(kwargs["opacity"] * 100)
        kwargs["st_mtime"] = kwargs["fstat"].st_mtime
        kwargs["st_size"] = kwargs["fstat"].st_size

        params = [
            "%(original_basename)s",
            "wm",
            "w%(watermark)i",
            "o%(opacity)i",
            "gs%(greyscale)i",
            "r%(rotation)i",
            "fm%(st_mtime)i",
            "fz%(st_size)i",
            "p%(position)s",
        ]

        scale = kwargs.get("scale", None)
        if scale and scale != mark.size:
            params.append("_s%i" % (float(kwargs["scale"][0]) / mark.size[0] * 100))

        if kwargs.get("tile", None):
            params.append("_tiled")

        # make thumbnail filename
        filename = "%s%s" % ("_".join(params), kwargs["ext"])

        return filename % kwargs

    def get_url_path(self, basedir, original_basename, ext, name, obscure=True):
        """Determines an appropriate watermark path"""

        try:
            hash = hashlib.sha1(smart_str(name)).hexdigest()
        except TypeError:
            hash = hashlib.sha1(smart_str(name).encode("utf-8")).hexdigest()

        # figure out where the watermark would be saved on the filesystem
        if obscure is True:
            logger.debug("Obscuring original image name: %s => %s" % (name, hash))
            url_path = os.path.join(basedir, hash + ext)
        else:
            logger.debug("Not obscuring original image name.")
            url_path = os.path.join(basedir, hash, original_basename + ext)

        # make sure the destination directory exists
        try:
            fpath = self._get_filesystem_path(url_path)
            os.makedirs(os.path.dirname(fpath))
        except OSError as e:
            if e.errno == errno.EEXIST:
                pass  # not to worry, directory exists
            else:
                logger.error("Error creating path: %s" % traceback.format_exc())
                raise
        else:
            logger.debug("Created directory: %s" % os.path.dirname(fpath))

        return url_path

    def create_watermark(self, target, mark, fpath, quality=QUALITY, **kwargs):
        """Create the watermarked image on the filesystem"""

        im = utils.watermark(target, mark, **kwargs)
        if not kwargs.get("noalpha", True) is False:
            im = im.convert("RGB")
        im.save(fpath, quality=quality)
        return im


@register.filter
def watermark(url, args=""):
    """
    Returns the URL to a watermarked copy of the image specified.

    """
    # initialize some variables
    args = args.split(",")

    params = dict(
        name=args.pop(0),
        opacity=0.5,
        tile=False,
        scale=1.0,
        greyscale=False,
        rotation=0,
        position=None,
        noalpha=True,
        quality=QUALITY,
        obscure=OBSCURE_ORIGINAL,
        random_position_once=RANDOM_POSITION_ONCE,
    )

    params["url"] = unquote(url)

    # iterate over all parameters to see what we need to do
    for arg in args:
        key, value = arg.split("=")
        key, value = key.strip(), value.strip()
        if key == "position":
            params["position"] = value
        elif key == "opacity":
            params["opacity"] = utils._percent(value)
        elif key == "tile":
            params["tile"] = bool(int(value))
        elif key == "scale":
            params["scale"] = value
        elif key == "greyscale":
            params["greyscale"] = bool(int(value))
        elif key == "rotation":
            params["rotation"] = value
        elif key == "noalpha":
            params["noalpha"] = bool(int(value))
        elif key == "quality":
            params["quality"] = int(value)
        elif key == "obscure":
            params["obscure"] = bool(int(value))
        elif key == "random_position_once":
            params["random_position_once"] = bool(int(value))

    return Watermarker()(**params)
